﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace HolePunchServer
{
    internal class Program
    {
        private Dictionary<string, IPEndPoint> _ipList = new Dictionary<string, IPEndPoint>();
        private int _port = 12800;

        private static void Main(string[] args)
        {
            Program pg = new Program();
            pg.Start();
        }

        private UdpClient _server;

        private void Start()
        {
            _server = new UdpClient(_port);
            Console.WriteLine("UDP Port: " + _port);
            _server.BeginReceive(udpReceiveCallback, _server);

            ThreadPool.QueueUserWorkItem((WaitCallback)pingCallback);

            Console.WriteLine("Press any key to exit...");
            System.Console.ReadLine();
        }

        private bool isFullClient = false;

        private void pingCallback(object state)
        {
            //if (_ipList.Count < 2)
            //    isFullClient = false;

            //if (isFullClient == true)
            //{
            //    Thread.Sleep(5000); // 5초에 한번씩 클라이언트 목록 전송
            //    ThreadPool.QueueUserWorkItem((WaitCallback)pingCallback);
            //    return;
            //}

            //if (_ipList.Count >= 2)
            //{
            //    isFullClient = true;
            //}

            lock (_ipList)
            {
                string txt = SerializeAsText(_ipList);
                byte[] bytes = Encoding.UTF8.GetBytes(txt);
                foreach (var item in _ipList)
                {
                    _server.BeginSend(bytes, bytes.Length, item.Value, udpSendCallback, _server);
                }
            }

            //if (_ipList.Count != 2)
            //{
            Thread.Sleep(5000); // 5초에 한번씩 클라이언트 목록 전송
            ThreadPool.QueueUserWorkItem((WaitCallback)pingCallback);
            //}
        }

        private string SerializeAsText(Dictionary<string, IPEndPoint> ipList)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("LIST$");
            foreach (var item in ipList)
            {
                sb.Append(item.Key);
                sb.Append(":" + item.Value.Address.ToString());
                sb.Append(":" + item.Value.Port.ToString());
                sb.Append(",");
            }

            return sb.ToString();
        }

        private void udpReceiveCallback(IAsyncResult ar)
        {
            try
            {
                UdpClient udpServer = ar.AsyncState as UdpClient;
                IPEndPoint remoteEndPoint = null;

                byte[] receiveBytes = null;

                try
                {
                    receiveBytes = udpServer.EndReceive(ar, ref remoteEndPoint);
                }
                catch (SocketException ex)
                {
                    Console.WriteLine(ex.ToString());
                }

                if (receiveBytes != null)
                {
                    string clientName = Encoding.UTF8.GetString(receiveBytes);

                    Console.WriteLine("Receive from " + clientName + " [" + remoteEndPoint.Address.ToString() + ":" + remoteEndPoint.Port + "]");

                    if (_ipList.ContainsKey(clientName) == false)
                    {
                        _ipList.Add(clientName, remoteEndPoint);
                    }
                    else
                    {
                        _ipList[clientName] = remoteEndPoint;
                    }

                    bool sent = false;
                    foreach (var item in _ipList.Keys)
                    {
                        if (_ipList[item].Address.ToString() != remoteEndPoint.Address.ToString())
                        {
                            byte[] sentBytes = Encoding.UTF8.GetBytes(_ipList[item].Port.ToString());
                            udpServer.BeginSend(sentBytes, sentBytes.Length, remoteEndPoint, udpSendCallback, udpServer);
                            sent = true;
                            break;
                        }
                    }

                    if (sent == false)
                    {
                        byte[] sentBytes = Encoding.UTF8.GetBytes("ACK$");
                        udpServer.BeginSend(sentBytes, sentBytes.Length, remoteEndPoint, udpSendCallback, udpServer);
                    }

                    OutputIPList();
                }

                udpServer.BeginReceive(udpReceiveCallback, udpServer);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        private void udpSendCallback(IAsyncResult ar)
        {
            try
            {
                UdpClient udpClient = ar.AsyncState as UdpClient;
                udpClient.EndSend(ar);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Send Error: " + ex.ToString());
            }
        }

        private void OutputIPList()
        {
            foreach (var item in _ipList.Keys)
            {
                IPEndPoint ipAddr = _ipList[item];
                Console.WriteLine(item + " == " + ipAddr.ToString() + ":" + ipAddr.Port);
            }
        }
    }
}